package org.randombits.confluence.metadata.impl;

import com.atlassian.cache.Cache;
import com.atlassian.cache.CacheManager;
import com.atlassian.confluence.core.ContentEntityManager;
import com.atlassian.confluence.core.ContentEntityObject;
import com.atlassian.confluence.core.ContentPropertyManager;
import com.atlassian.confluence.core.VersionHistorySummary;
import com.atlassian.plugin.PluginAccessor;
import com.atlassian.plugin.event.PluginEventManager;
import com.atlassian.plugin.tracker.DefaultPluginModuleTracker;
import com.atlassian.plugin.tracker.PluginModuleTracker;
import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.core.BaseException;
import org.apache.log4j.Logger;
import org.randombits.confluence.metadata.*;
import org.randombits.confluence.metadata.impl.handler.DirectTypeHandler;
import org.randombits.confluence.metadata.xstream.MetadataXStream;
import org.randombits.storage.Aliasable;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.annotation.Qualifier;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.*;

public class DefaultMetadataManager implements MetadataManager, DisposableBean {

    private static final Logger LOG = Logger.getLogger( MetadataManager.class );

    private static final String METADATA_PREFIX = "~metadata";

    private static final String METADATA_CACHE_KEY = MetadataManager.class.getName();

    private final CacheManager cacheManager;

    private final ContentPropertyManager contentPropertyManager;

    private final ContentEntityManager contentEntityManager;

    private final PluginModuleTracker<TypeHandler, TypeHandlerModuleDescriptor> typeHandlerTracker;

    private XStream xStream;

    private List<TypeHandler> typeHandlers;

    public DefaultMetadataManager( CacheManager cacheManager, PluginAccessor pluginAccessor, PluginEventManager pluginEventManager, ContentPropertyManager contentPropertyManager, @Qualifier("contentEntityManager") ContentEntityManager contentEntityManager ) {
        this.cacheManager = cacheManager;
        this.contentPropertyManager = contentPropertyManager;
        this.contentEntityManager = contentEntityManager;

        typeHandlers = new ArrayList<TypeHandler>();

        xStream = new MetadataXStream();
        xStream.setClassLoader( getClass().getClassLoader() );
        applyAliases( this );

        initTypeHandlers();

        typeHandlerTracker = new DefaultPluginModuleTracker<TypeHandler, TypeHandlerModuleDescriptor>( pluginAccessor, pluginEventManager, TypeHandlerModuleDescriptor.class,
                new PluginModuleTracker.Customizer<TypeHandler, TypeHandlerModuleDescriptor>() {
                    @Override
                    public TypeHandlerModuleDescriptor adding( TypeHandlerModuleDescriptor moduleDescriptor ) {
                        addTypeHandler( moduleDescriptor.getModule() );
                        return moduleDescriptor;
                    }

                    @Override
                    public void removed( TypeHandlerModuleDescriptor moduleDescriptor ) {
                        removeTypeHandler( moduleDescriptor.getModule() );
                    }
                } );
    }

    protected void initTypeHandlers() {
        addDirectTypeHandler(
                String.class, Character.class, char.class,
                Byte.class, byte.class, Short.class, short.class, Integer.class, int.class, Long.class, long.class,
                Float.class, float.class, Double.class, double.class, BigDecimal.class, BigInteger.class,
                Boolean.class, boolean.class,
                Date.class );
    }

    private void addDirectTypeHandler( Class<?>... type ) {
        addTypeHandler( new DirectTypeHandler( type ) );
    }

    public void addTypeHandler( TypeHandler handler ) {
        typeHandlers.add( handler );
        if ( handler instanceof HasAlias )
            ( (HasAlias) handler ).applyAliases( this );
    }

    public void removeTypeHandler( TypeHandler handler ) {
        typeHandlers.remove( handler );
    }

    /**
     * Returns the provided 'storable' value as the original value.
     *
     * @param storable The stored value.
     * @return The original, or <code>null</code> if it is not supported.
     */
    public Object fromStorable( Object storable ) throws TypeConversionException {
        if ( storable == null )
            return null;

        for ( TypeHandler handler : typeHandlers ) {
            if ( handler.supportsStorable( storable ) )
                return handler.getOriginal( storable );
        }
        return null;
    }

    public Object toStorable( Object original ) throws TypeConversionException {
        if ( original == null )
            return null;

        for ( TypeHandler handler : typeHandlers ) {
            if ( handler.supportsOriginal( original ) )
                return handler.getStorable( original );
        }
        return null;
    }

    public void applyAliases( XStream xStream ) {
        applyAliases( new AliasableXStream( xStream ) );
    }

    /**
     * Applies the standard metadata aliases to the {@link org.randombits.storage.Aliasable} instance.
     *
     * @param aliasable The {@link org.randombits.storage.Aliasable} instance to alias.
     */
    public void applyAliases( Aliasable aliasable ) {
        // Alias the standard reference classes.
        for ( TypeHandler handler : typeHandlers ) {
            if ( handler instanceof HasAlias )
                ( (HasAlias) handler ).applyAliases( aliasable );
        }
    }

    /**
     * Returns a content property value for the specified content/key combo.
     *
     * @param content The content object.
     * @param key     The key to retrieve.
     * @return The property value.
     */
    private String getProperty( ContentEntityObject content, String key ) {
        return contentPropertyManager.getTextProperty( content, key );
    }

    private ContentEntityObject findVersion( ContentEntityObject content, int version ) {
        if ( content.getVersion() == version )
            return content;

        if ( !content.isLatestVersion() )
            content = (ContentEntityObject) content.getLatestVersion();

        List<VersionHistorySummary> summaries = contentEntityManager.getVersionHistorySummaries( content );
        for ( VersionHistorySummary summary : summaries ) {
            if ( summary.getVersion() == version )
                return contentEntityManager.getById( summary.getId() );
        }

        return null;
    }

    private Cache getCache() {
        return cacheManager.getCache( METADATA_CACHE_KEY );
    }

    /**
     * Loads a read-only, cached copy of the current data for the specified
     * content. This is much more efficient than
     * {@link #loadWritableData(ContentEntityObject)} when you only need read
     * access to the current data.
     *
     * @param content The content object.
     * @return The current metadata.
     */
    public MetadataStorage loadReadableData( ContentEntityObject content ) {
        Map<String, Object> data = null;

        if ( content != null ) {
            Long id = content.getId();
            data = (Map<String, Object>) getCache().get( id );

            if ( data == null ) {
                data = loadDataMap( content );

                Cache readThroughCache = getCache();
                readThroughCache.put( id, data );
            }
        }

        MetadataStorage metadata;
        if ( data != null ) {
            metadata = new StandardMetadataStorage( this, content, data, true );
        } else {
            metadata = new EmptyMetadataStorage();
        }

        return metadata;
    }

    /**
     * Loads the data for the specified content object. If the version number is
     * not valid, <code>null</code> is returned.
     *
     * @param content The content object.
     * @param version The version to retrieve.
     * @return the data for the specified version of the content object.
     */
    public MetadataStorage loadWritableData( ContentEntityObject content, int version ) {
        content = findVersion( content, version );
        if ( content != null )
            return loadWritableData( content );
        return null;
    }

    /**
     * Loads the data into a MetadataStorage instance. The storage will load any
     * legacy data, if it exists.
     *
     * @param content The content object.
     * @return The metadata storage.
     */
    public MetadataStorage loadWritableData( ContentEntityObject content ) {
        MetadataStorage metadata = null;

        if ( content != null ) {
            Map<String, Object> data = loadDataMap( content );

            if ( data == null )
                data = new HashMap<String, Object>();

            metadata = new StandardMetadataStorage( this, content, data, false );
        } else {
            metadata = new EmptyMetadataStorage();
        }

        return metadata;
    }

    private Map<String, Object> loadDataMap( ContentEntityObject content ) {
        // look for the data.
        Map<String, Object> data = null;
        String xml = loadDataXML( content );

        if ( xml != null ) {
            try {
                data = (Map<String, Object>) xStream.fromXML( xml );
            } catch ( BaseException e ) {
                LOG.error( "Conversion issue while loading metadata for content: " + content + "; version: "
                        + content.getVersion(), e );
                data = new MetadataMap();
            }

        }
        return data;
    }

    /**
     * Loads the data for the specified content object, returning it as an XML
     * string. This method will not load any legacy data.
     *
     * @param content The content object.
     * @return The XML String, or <code>null</code> if no data has been saved.
     */
    public String loadDataXML( ContentEntityObject content ) {
        int version = content.getVersion();
        ContentEntityObject latest = (ContentEntityObject) content.getLatestVersion();

        for ( int i = version; i >= 1; i-- ) {
            String xml = getProperty( latest, METADATA_PREFIX + "." + i );
            if ( xml != null ) {
                return xml;
            }
        }

        // If we get this far, we couldn't find any metadata.
        return null;
    }

    /**
     * Saves the data as the next version. This is useful for pre-saving data
     * when other processes will manage the actual creation of the next content
     * version.
     *
     * @param data The data to save.
     */
    public void saveNextData( MetadataStorage data ) {
        this.saveData( data, false );
    }

    /**
     * Saves the metadata. If <code>overwriteCurrent</code> is
     * <code>true</code>, the current set of existing data will be
     * overwritten. If not, it will be saved to a version one number greater
     * than the current version for the content on the assumption that the
     * content object will be saved to a new version immediately afterwards.
     *
     * @param data                  The metadata to save.
     * @param currentContentVersion If <code>true</code> the previous version will be
     *                              overwritten.
     */
    public void saveData( MetadataStorage data, boolean currentContentVersion ) {
        ContentEntityObject content = data.getContent();

        if ( content != null ) {
            ContentEntityObject latest = (ContentEntityObject) content.getLatestVersion();
            int version = latest.getVersion();
            if ( version == 0 || !currentContentVersion )
                version++;

            MetadataMap map = toMetadataMap( data.getBaseMap() );
            String xml = xStream.toXML( map );

            ContentPropertyManager cpm = contentPropertyManager;

            cpm.setTextProperty( latest, METADATA_PREFIX + "." + version, xml );

            // Reset the cache
            clearCache( latest );
        }
    }

    /**
     * Clears the cache for the specified content.
     *
     * @param content The content to clear the metadata cache for.
     */
    private void clearCache( ContentEntityObject content ) {
        Cache readThroughCache = getCache();
        Long id = content.getId();
        readThroughCache.remove( id );
    }

    /**
     * By default, the metadata manager will cache up to 1000 data items for 60
     * minutes. This will clear any cached data from memory.
     */
    public void clearCache() {
        getCache().removeAll();
    }

    /**
     * Clears the data storage area for the next version of the content object.
     * This is useful when data has been pre-saved and the overall page-save
     * fails.
     *
     * @param content The content object.
     */
    public void clearNextData( ContentEntityObject content ) {
        if ( !content.isLatestVersion() )
            content = (ContentEntityObject) content.getLatestVersion();

        contentPropertyManager.setTextProperty( content, METADATA_PREFIX + "."
                + ( content.getVersion() + 1 ), null );
    }

    /**
     * Converts a regular map structure into a MetadataMap structure for
     * storage.
     *
     * @param baseMap The Map to convert.
     * @return the map as a {@link MetadataMap}.
     */
    private MetadataMap toMetadataMap( Map<String, ?> baseMap ) {
        if ( baseMap == null )
            return null;

        if ( baseMap instanceof MetadataMap )
            return (MetadataMap) baseMap;

        MetadataMap metadataMap = new MetadataMap();

        for ( Map.Entry<String, ?> e : baseMap.entrySet() ) {
            Object value = e.getValue();
            if ( value instanceof Map )
                metadataMap.put( e.getKey(), toMetadataMap( (Map<String, ?>) value ) );
            else
                metadataMap.put( e.getKey(), e.getValue() );
        }

        return metadataMap;
    }

    /**
     * Adds an alias for the specified class. Aliases are used in the
     * serialization/deserialization process.
     *
     * @param alias        The alias to add.
     * @param aliasedClass The class being aliased.
     */
    public void addAlias( String alias, Class<?> aliasedClass ) {
        xStream.alias( alias, aliasedClass );
    }

    /**
     * Returns a read-only {@link MetadataStorage} instance based on the
     * provided {@link Map}. Trying to save the map will fail, since it is not
     * associated with a particular piece of content.
     *
     * @param map The map to use.
     * @return The {@link MetadataStorage} instance.
     */
    public MetadataStorage asMetadataStorage( Map<String, Object> map ) {
        return new StandardMetadataStorage( this, null, map, true );
    }


    public boolean isStorable( Object value ) {
        for ( TypeHandler handler : typeHandlers ) {
            if ( handler.supportsStorable( value ) )
                return true;
        }
        return false;
    }

    @Override
    public String toXML( Object value ) throws TypeConversionException {
        value = toStorable( value );
        if ( value != null ) {
            try {
                return xStream.toXML( value );
            } catch ( Exception e ) {
                throw new TypeConversionException( e.getMessage(), e );
            }
        } else {
            return null;
        }
    }

    @Override
    public Object fromXML( String xml ) throws TypeConversionException {
        if ( xml != null ) {
            try {
                Object value = xStream.fromXML( xml );
                return fromStorable( value );
            } catch ( Exception e ) {
                throw new TypeConversionException( e.getMessage(), e );
            }
        }
        return null;
    }

    private static class AliasableXStream implements Aliasable {

        XStream xStream;

        AliasableXStream( XStream xStream ) {
            this.xStream = xStream;
        }

        /**
         * Adds the specified alias for the class.
         *
         * @param alias        The alias.
         * @param aliasedClass The class to alias.
         */
        public void addAlias( String alias, Class<?> aliasedClass ) {
            xStream.alias( alias, aliasedClass );
        }
    }

    @Override
    public void destroy() throws Exception {
        typeHandlers.clear();
        typeHandlerTracker.close();
    }
}
